
`timescale 1 ns/ 1 ps

module LVDS_7to1_TX (
      CLK_Tx,
      RST_Tx,
              
      T0_in,
      T1_in,
      T2_in,
      T3_in,
              
      TCLK_out,
      T0_out,
      T1_out,
      T2_out,
      T3_out 
    );

   input        CLK_Tx;   // Tx module input clock 	// 40 Mhz
   input        RST_Tx;   // Tx module reset
     
   input  [6:0] T0_in;    // 7-bit parallel data
   input  [6:0] T1_in;    // 7-bit parallel data
   input  [6:0] T2_in;    // 7-bit parallel data
   input  [6:0] T3_in;    // 7-bit parallel data
   
   output       TCLK_out; // LVDS clock output pair 
   output       T0_out;   // LVDS data output pair 0
   output       T1_out;   // LVDS data output pair 1
   output       T2_out;   // LVDS data output pair 2
   output       T3_out;   // LVDS data output pair 3
         

   wire [6:0] txdata_in  [0:3];
   wire [3:0] txdata_odd [0:3];

   wire [3:0] txclk_odd;
   wire [3:0] tx_do;

   wire [2:0] tx_wr_state;
   wire [3:0] tx_rd_state;      
         
   wire eclk;
   wire sclk;

   reg  sync_lock;
   wire tx_pll_lock;
   wire tx_reset;

   
   txpll PLL_INST
         (
          .CLK   (CLK_Tx),				//  40 Mhz
          .RESET (RST_Tx),
          .CLK_3p5x (eclk),				// 140 Mhz
          .CLK_1p75x (sclk),			//  70 Mhz
          .LOCK  (tx_pll_lock)
         );

   always @(negedge tx_pll_lock or posedge CLK_Tx)
   begin
      if (tx_pll_lock == 1'b0)
         sync_lock <= 1'b0;
      else
         sync_lock <= 1'b1;
   end

   assign txpll_lock  = sync_lock; 
   assign tx_reset = ~sync_lock;

      ser_fsm ser_fsm_inst
              (
               .TXCLK   (CLK_Tx),
               .CLK     (sclk),
               .RST     (tx_reset),
               .tx_wr_state (tx_wr_state),
               .tx_rd_state (tx_rd_state)
              );
         
   //--------------------------------------------------------------------
   //-- Data pair 0 ~ 3
   //--------------------------------------------------------------------
   
      assign txdata_in[0] = T0_in;
      assign txdata_in[1] = T1_in;
      assign txdata_in[2] = T2_in;
      assign txdata_in[3] = T3_in;
 
   
   genvar i;
   
   generate
   for (i=0; i < 4; i=i+1)
   begin : u
      serializer ser_inst 
                 (
                  .TXCLK   (CLK_Tx),
                  .CLK     (sclk),
                  .DATAIN  (txdata_in[i]),
                  .RST     (tx_reset),
                  .tx_wr_state (tx_wr_state),
                  .tx_rd_state (tx_rd_state),         
                  .DATAOUT (txdata_odd[i])
                 );

      tx_oddr_x2 txdata_inst
                 (      
				  .RST (tx_reset),
                  .EClk (eclk),
                  .SClk (sclk),
                  .Data (txdata_odd[i]),
                  .Q (tx_do[i])
                 );
    end
    endgenerate


   assign T0_out = tx_do[0];
   assign T1_out = tx_do[1];
   assign T2_out = tx_do[2];
   assign T3_out = tx_do[3];

   //--------------------------------------------------------------------
   //-- Clock pair
   //--------------------------------------------------------------------

   // Note that using "1100011" for generating the clock will make a
   // "2-bit shifting" between the clock and data pairs.  This shift
   // is required for compatibility with National Semiconductor's
   // CameraLink LVDS chips.
   //             __________________                 __________________
   // Clock _____/           1    1 \__0____0____0__/ 1    1           \
   //        ___  ___  ___  ___  ___  ___  ___  ___  ___  ___  ___  ___ 
   // Data  /   \/   \/   \/bit\/bit\/bit\/bit\/bit\/bit\/bit\/   \/   \
   //       \___/\___/\___/\_6_/\_5_/\_4_/\_3_/\_2_/\_1_/\_0_/\___/\___/
   //

   serializer_txclk ser_clk_inst
             (
              .CLK     (sclk),
              .DATAIN  (7'b1100011),
              .RST     (tx_reset),
              .tx_rd_state (tx_rd_state),      
              .DATAOUT (txclk_odd)
             );

   tx_oddr_x2 txdata_clk_inst
              (   
			   .RST (tx_reset),
               .EClk (eclk),
               .SClk (sclk),
               .Data (txclk_odd),
               .Q (TCLK_out)
              );

endmodule
